{-|
Module      : Gargantext.Core.AsyncUpdates
Description : Asynchronous updates to the frontend
Copyright   : (c) CNRS, 2024-Present
License     : AGPL + CECILL v3
Maintainer  : team@gargantext.org
Stability   : experimental
Portability : POSIX
-}

module Gargantext.Core.AsyncUpdates
where

import Gargantext.Core.Types (NodeId, UserId)
import Protolude


{-

I imagine the workflow as follows:
- somewhere in the code (or in the async job worker) we decide to send
  an update message to all interested users
- such an action (UserAction) can be associated with the triggering
  user (but doesn't have to be)
- we compute interested users for given notification
- we broadcast (using our broker) these notifications to all
  interested users 
- the broadcast message is either simple (meaning: hey, we have new
  data, if you want you can send an update request) or we could send
  the changed data already

On the client side it looks more or less like this:
- specific components decide to subscribe to specific
  UserNotifications: task component is interested in running tasks (for
  given node id), tree component is interested in the tree and its
  first-level children (same for the children components)
- the components react to events accordingly (usually by pulling in
  new data)

Thus, for example, the triple (user_id, node_id, "update_tree")
defines a "update tree for given user and given node" subscription to
this event, both for server and client. This triple is then the
"touching point" between client and server. Through that point, update
messages are sent from server.

Subscription to topics is important IMHO because it allows to target
clients directly rather than broadcasting messages to everyone. This
reduces latency and is more secure. At the same time it is a bit more
complicated because we need to agree on the topic schema both on
server and client.

-}
    

-- | Various update actions
data UpdateAction =
  -- | Update given Servant Job (we currently send a request every
  -- | second to get job status).
  --  UpdateJob JobID
  -- | Given parent node id, trigger update of the node and its
  --   children (e.g. list is automatically created in a corpus)
  UpdateTree NodeId
  deriving (Eq, Show)

data UserSource =
    USUser UserId
  | USSystem
  deriving (Eq, Show)
    
-- | Action possibly associated with user who triggered it (there can
--   be system actions as well)
data UserAction =
  UserAction UserSource UpdateAction
  deriving (Eq, Show)
    
-- | Represents a notification that goes to a given user. This is
--   directly sent via WebSockets.
data UserNotification =
  UserNotification UserId UserAction
  deriving (Eq, Show)

-- | What we want now is, given a UserAction action, generate all
--   interested users to which the notification will be sent.
--   This function lives in a monad because we have to fetch users
--   from DB.
notificationsForUserAction :: UserAction -> m [ UserNotification ]
notificationsForUserAction = undefined


-- | Stores connection type associated with given user.
--   We probably should set conn = Servant.API.WebSocket.Connection
data ConnectedUser conn =
  ConnectedUser UserId conn


-- | Given a UserNotification and all connected users, send it to
--   interested ones. Possibly we could make this function as part of
--   a typeclass so that we can decide how to send the notification
--   based on whether we choose pure WebSockets, NATS or something
--   else.
sendNotification :: UserNotification -> [ ConnectedUser conn ] -> m ()
sendNotification = undefined
